import random
import typing, math
from dataclasses import dataclass
import terminaltexteffects.utils.easing as easing

if typing.TYPE_CHECKING:
    from terminaltexteffects import base_character


@dataclass(eq=True, frozen=True)
class Coord:
    """A coordinate with row and column values.

    Args:
        column (int): column value
        row (int): row value"""

    column: int
    row: int


@dataclass
class Waypoint:
    """A coordinate, speed, and easing function.

    Args:
        waypoint_id (str): unique identifier for the waypoint
        coord (Coord): coordinate
        bezier_control (Coord | None): coordinate of the control point for a bezier curve. Defaults to None.
    """

    waypoint_id: str
    coord: Coord
    bezier_control: tuple[Coord, ...] | Coord | None = None

    def __post_init__(self) -> None:
        if self.bezier_control and isinstance(self.bezier_control, Coord):
            self.bezier_control = (self.bezier_control,)

    def __eq__(self, other: typing.Any) -> bool:
        if not isinstance(other, Waypoint):
            return NotImplemented
        return self.coord == other.coord

    def __hash__(self):
        return hash(self.waypoint_id)


@dataclass
class Segment:
    """A segment of a path consisting of two waypoints and the distance between them.

    Args:
        start (Waypoint): start waypoint
        end (Waypoint): end waypoint
        distance (float): distance between the start and end waypoints
    """

    start: Waypoint
    end: Waypoint
    distance: float

    def __post_init__(self) -> None:
        self.enter_event_triggered: bool = False
        self.exit_event_triggered: bool = False

    def get_coord_on_segment(self, distance_factor: float) -> Coord:
        """Returns the coordinate at the given distance along the segment.

        Args:
            distance_factor (float): distance factor

        Returns:
            Coord: Coordinate at the given distance.
        """
        if self.start.bezier_control:
            return Motion.find_coord_on_bezier_curve(
                self.start.coord,
                self.start.bezier_control,
                self.end.coord,
                distance_factor,
            )
        else:
            return Motion.find_coord_on_line(self.start.coord, self.end.coord, distance_factor)

    def __eq__(self, other: typing.Any) -> bool:
        if not isinstance(other, Segment):
            return NotImplemented
        return self.start == other.start and self.end == other.end

    def __hash__(self):
        return hash((self.start, self.end))


@dataclass
class Path:
    """
    Represents a path consisting of multiple waypoints for motion.

    Attributes:
        path_id (str): The unique identifier for the path.
        speed (float): speed
        ease (Callable | None): easing function for character movement. Defaults to None.
        layer (int | None): layer to move the character to, if None, layer is unchanged. Defaults to None.
        hold_time (int): number of animation steps to hold the character at the end of the path. Defaults to 0.
        loop (bool): Whether the path should loop back to the beginning. Default is False.
    """

    path_id: str
    speed: float = 1.0
    ease: typing.Callable | None = None
    layer: int | None = None
    hold_time: int = 0
    loop: bool = False

    def __post_init__(self) -> None:
        """
        Initializes the Path object and calculates the total distance and maximum steps.
        """
        self.segments: list[Segment] = []
        self.waypoints: list[Waypoint] = []
        self.waypoint_lookup: dict[str, Waypoint] = {}
        self.total_distance: float = 0
        self.current_step: int = 0
        self.max_steps: int = 0
        self.hold_time_remaining = self.hold_time
        self.last_distance_reached: float = 0  # used for animation syncing to distance
        self.origin_segment: Segment | None = None

    def new_waypoint(
        self,
        waypoint_id: str,
        coord: Coord,
        *,
        bezier_control: tuple[Coord, ...] | Coord | None = None,
    ) -> Waypoint:
        """Creates a new Waypoint and appends adds it to the Path.

        Args:
            waypoint_id (str): unique identifier for the waypoint
            coord (Coord): coordinate
            bezier_control (tuple[Coord, ...] | Coord | None): coordinate of the control point for a bezier curve. Defaults to None.

        Returns:
            Waypoint: The new waypoint.
        """
        new_waypoint = Waypoint(waypoint_id, coord, bezier_control=bezier_control)
        self._add_waypoint_to_path(new_waypoint)
        return new_waypoint

    def _add_waypoint_to_path(self, waypoint: Waypoint) -> None:
        """Adds a waypoint to the path and updates the total distance and maximum steps.

        Args:
            waypoint (Waypoint): waypoint to add
        """
        self.waypoint_lookup[waypoint.waypoint_id] = waypoint
        self.waypoints.append(waypoint)
        if len(self.waypoints) < 2:
            return

        if waypoint.bezier_control:
            distance_from_previous = Motion.find_length_of_bezier_curve(
                self.waypoints[-2].coord, waypoint.bezier_control, waypoint.coord
            )
        else:
            distance_from_previous = Motion.find_length_of_line(
                self.waypoints[-2].coord,
                waypoint.coord,
            )
        self.total_distance += distance_from_previous
        self.segments.append(Segment(self.waypoints[-2], waypoint, distance_from_previous))
        self.max_steps = round(self.total_distance / self.speed)

    def step(self, event_handler: "base_character.EventHandler") -> Coord:
        """Progresses to the next step along the path and returns the coordinate at that step."""
        if not self.max_steps or self.current_step >= self.max_steps or not self.total_distance:
            # if the path has zero distance or there are no more steps, return the coordinate of the final waypoint in the path
            return self.segments[-1].end.coord
        else:
            self.current_step += 1
        if self.ease:
            distance_factor = self.ease(self.current_step / self.max_steps)
        else:
            distance_factor = self.current_step / self.max_steps

        distance_to_travel = distance_factor * self.total_distance
        self.last_distance_reached = distance_to_travel
        for segment in self.segments:
            if distance_to_travel <= segment.distance:
                active_segment = segment
                if not segment.enter_event_triggered:
                    segment.enter_event_triggered = True
                    event_handler.handle_event(event_handler.Event.SEGMENT_ENTERED, segment.end)
                break
            distance_to_travel -= segment.distance
            if not segment.exit_event_triggered:
                segment.exit_event_triggered = True
                event_handler.handle_event(event_handler.Event.SEGMENT_EXITED, segment.end)
        else:  # if the distance_to_travel is further than the last waypoint, preserve the distance from the start of the final segment
            active_segment = self.segments[-1]
            distance_to_travel += active_segment.distance

        segment_distance_to_travel_factor = distance_to_travel / active_segment.distance

        if active_segment.end.bezier_control:
            next_coord = Motion.find_coord_on_bezier_curve(
                active_segment.start.coord,
                active_segment.end.bezier_control,
                active_segment.end.coord,
                segment_distance_to_travel_factor,
            )
        else:
            next_coord = Motion.find_coord_on_line(
                active_segment.start.coord, active_segment.end.coord, segment_distance_to_travel_factor
            )

        return next_coord

    def __eq__(self, other: typing.Any) -> bool:
        if not isinstance(other, Path):
            return NotImplemented
        return self.path_id == other.path_id

    def __hash__(self):
        return hash(self.path_id)


class Motion:
    def __init__(self, character: "base_character.EffectCharacter"):
        """Motion class for managing the movement of a character.

        Args:
            character (base_character.EffectCharacter): The EffectCharacter to move.
        """
        self.paths: dict[str, Path] = {}
        self.character = character
        self.current_coord: Coord = Coord(character.input_coord.column, character.input_coord.row)
        self.previous_coord: Coord = Coord(-1, -1)
        self.active_path: Path | None = None

    @staticmethod
    def find_coords_on_circle(origin: Coord, radius: int, num_points: int) -> list[Coord]:
        """Finds points on a circle.

        Args:
            origin (Coord): origin of the circle
            radius (int): radius of the circle
            num_points (int): number of points to find

        Returns:
            list (Coord): list of Coord points on the circle
        """
        points = []
        for i in range(num_points):
            angle = 2 * math.pi * i / num_points
            x = origin.column + radius * math.cos(angle)
            # correct for terminal character height/width ratio by doubling the x distance from origin
            x_diff = x - origin.column
            x += x_diff
            y = origin.row + radius * math.sin(angle)
            points.append(Coord(round(x), round(y)))
        return points

    @staticmethod
    def find_coords_in_circle(origin: Coord, radius: int) -> list[Coord]:
        """Finds points that fall within a circle with the given origin and radius. Duplicate points
        after rounding are removed.

        Args:
            origin (Coord): origin of the circle
            radius (int): radius of the circle
            num_points (int): number of points to find

        Returns:
            list[Coord]: list of Coord points in the circle
        """
        points: list[Coord] = []
        for i in range(1, radius + 1):
            points.extend(set((Motion.find_coords_on_circle(origin, i, 7 * radius))))
        points = list(set(points))
        return points

    @staticmethod
    def find_coords_in_rect(origin: Coord, distance: int) -> list[Coord]:
        """Find coords that fall within a rectangle with the given origin and distance
        from the origin. Distance specifies the number of units in each direction from the origin.
        Final width = 2 * distance + 1, final height = 2 * distance + 1.

        Args:
            origin (Coord): center of the rectangle
            distance (int): distance from the origin

        Returns:
            list[Coord]: list of Coord points in the rectangle
        """
        left_boundary = origin.column - distance
        right_boundary = origin.column + distance
        top_boundary = origin.row - distance
        bottom_boundary = origin.row + distance
        coords: list[Coord] = []
        for column in range(left_boundary, right_boundary + 1):
            for row in range(top_boundary, bottom_boundary + 1):
                coords.append(Coord(column, row))

        return coords

    @staticmethod
    def find_coord_at_distance(origin: Coord, target: Coord, distance: float) -> Coord:
        """Finds the coordinate at the given distance along the line defined by the origin and target coordinates.

        Args:
            origin (Coord): origin coordinate (a)
            target (Coord): target coordinate (b)
            distance (float): distance from the target coordinate (b), away from the origin coordinate (a)

        Returns:
            Coord: Coordinate at the given distance (c).
        """
        total_distance = Motion.find_length_of_line(origin, target) + distance
        if total_distance == 0:
            return origin
        t = total_distance / Motion.find_length_of_line(origin, target)
        next_column, next_row = (
            ((1 - t) * origin.column + t * target.column),
            ((1 - t) * origin.row + t * target.row),
        )
        return Coord(round(next_column), round(next_row))

    @staticmethod
    def find_coord_on_bezier_curve(start: Coord, control: tuple[Coord, ...] | Coord, end: Coord, t: float) -> Coord:
        """
        Finds points on a quadratic or cubic bezier curve.

        Args:
            start (Coord): The starting coordinate of the curve.
            control (tuple[Coord, ...] | Coord): The control point(s) of the curve.
                For a quadratic bezier curve, a single control point is expected.
                For a cubic bezier curve, two control points are expected.
            end (Coord): The ending coordinate of the curve.
            t (float): The parameter value between 0 and 1 that determines the position on the curve.

        Returns:
            Coord: The coordinate on the bezier curve corresponding to the given parameter value.
        """
        if isinstance(control, Coord):
            control = (control,)
        if len(control) == 1:
            control1 = control[0]
            x = (1 - t) ** 2 * start.column + 2 * (1 - t) * t * control1.column + t**2 * end.column
            y = (1 - t) ** 2 * start.row + 2 * (1 - t) * t * control1.row + t**2 * end.row
        elif len(control) == 2:
            control1, control2 = control
            x = (
                (1 - t) ** 3 * start.column
                + 3 * (1 - t) ** 2 * t * control1.column
                + 3 * (1 - t) * t**2 * control2.column
                + t**3 * end.column
            )
            y = (
                (1 - t) ** 3 * start.row
                + 3 * (1 - t) ** 2 * t * control1.row
                + 3 * (1 - t) * t**2 * control2.row
                + t**3 * end.row
            )
        return Coord(round(x), round(y))

    @staticmethod
    def find_coord_on_line(start: Coord, end: Coord, t: float) -> Coord:
        """
        Finds points on a line.

        Args:
            start (Coord): The starting coordinate of the line.
            end (Coord): The ending coordinate of the line.
            t (float): The parameter value between 0 and 1 representing the position on the line.

        Returns:
            Coord: The coordinate on the line corresponding to the given parameter value.
        """
        x = (1 - t) * start.column + t * end.column
        y = (1 - t) * start.row + t * end.row
        return Coord(round(x), round(y))

    @staticmethod
    def find_length_of_bezier_curve(start: Coord, control: tuple[Coord, ...] | Coord, end: Coord) -> float:
        """
        Finds the length of a quadratic or cubic bezier curve.

        Args:
            start (Coord): The starting coordinate of the curve.
            control (tuple[Coord, ...] | Coord): The control point(s) of the curve.
            end (Coord): The ending coordinate of the curve.

        Returns:
            float: The length of the bezier curve.
        """
        length = 0.0
        prev_coord = start
        for t in range(1, 10):
            coord = Motion.find_coord_on_bezier_curve(start, control, end, t / 10)
            length += Motion.find_length_of_line(prev_coord, coord)
            prev_coord = coord
            prev_coord = coord
        return length

    @staticmethod
    def find_length_of_line(coord1: Coord, coord2: Coord) -> float:
        """Returns the length of the line intersecting coord1 and coord2.

        Args:
            coord1 (Coord): first coordinate
            coord2 (Coord): second coordinate

        Returns:
            float: length of the line
        """
        column_diff = coord2.column - coord1.column
        row_diff = coord2.row - coord1.row
        return math.hypot(column_diff, row_diff)

    def set_coordinate(self, coord: Coord) -> None:
        """Sets the current coordinate to the given coordinate.

        Args:
            coord (Coord): coordinate
        """
        self.current_coord = coord

    def new_path(
        self,
        path_id: str,
        *,
        speed: float = 1,
        ease: typing.Callable | None = None,
        layer: int | None = None,
        hold_time: int = 0,
        loop: bool = False,
    ) -> Path:
        """Creates a new Path and adds it to the Motion.paths dictionary with the path_id as key.

        Args:
            path_id (str): unique identifier for the path
            speed (float): speed
            ease (Callable | None): easing function for character movement. Defaults to None.
            layer (int | None): layer to move the character to, if None, layer is unchanged. Defaults to None.
            hold_time (int): number of animation steps to hold the character at the end of the path. Defaults to 0.
            loop (bool): whether the path should loop. Defaults to False.

        Returns:
            Path: The new path.
        """
        if path_id in self.paths:
            raise ValueError(f"Path with id {path_id} already exists.")
        new_path = Path(path_id, speed, ease, layer, hold_time, loop)
        self.paths[path_id] = new_path
        return new_path

    def movement_is_complete(self) -> bool:
        """Returns whether the character has reached the final coordinate and moved the requisite number of steps.

        Returns:
            bool: True if the character has reached the final coordinate and has taken the maximum number of steps, False otherwise.
        """
        if self.active_path is None:
            return True
        return False

    def _get_easing_factor(self, easing_func: typing.Callable) -> float:
        """Returns the percentage of total distance that should be moved based on the easing function.

        Args:
            easing_func (Callable): The easing function to use.

        Returns:
            float: The percentage of total distance to move.
        """
        if not self.active_path:
            raise ValueError("No active path.")
        elapsed_step_ratio = self.active_path.current_step / self.active_path.max_steps
        return easing_func(elapsed_step_ratio)

    def chain_paths(self, paths: list[Path], loop=False):
        """Creates a chain of paths by registering activation events for each path such
        that paths[n] activates paths[n+1] when reached. If loop is True, paths[-1] activates
        paths[0] when reached.

        Args:
            paths (list[Path]): list of paths to chain
            loop (bool, optional): Whether the chain should loop. Defaults to False.
        """
        if len(paths) < 2:
            return
        for i, path in enumerate(paths):
            if i == 0:
                continue
            self.character.event_handler.register_event(
                self.character.event_handler.Event.PATH_COMPLETE,
                paths[i - 1],
                self.character.event_handler.Action.ACTIVATE_PATH,
                path,
            )
        if loop:
            self.character.event_handler.register_event(
                self.character.event_handler.Event.PATH_COMPLETE,
                paths[-1],
                self.character.event_handler.Action.ACTIVATE_PATH,
                paths[0],
            )

    def activate_path(self, path: Path) -> None:
        """Activates the first waypoint in the path.

        Args:
            path (Path): the Path to activate
        """
        self.active_path = path
        first_waypoint = self.active_path.waypoints[0]
        if first_waypoint.bezier_control:
            distance_to_first_waypoint = self.find_length_of_bezier_curve(
                self.current_coord, first_waypoint.bezier_control, first_waypoint.coord
            )
        else:
            distance_to_first_waypoint = self.find_length_of_line(
                self.current_coord,
                first_waypoint.coord,
            )
        self.active_path.total_distance += distance_to_first_waypoint
        if self.active_path.origin_segment:
            self.active_path.segments.pop(0)
            self.active_path.total_distance -= self.active_path.origin_segment.distance
        self.active_path.origin_segment = Segment(
            Waypoint("origin", self.current_coord), first_waypoint, distance_to_first_waypoint
        )
        self.active_path.segments.insert(0, self.active_path.origin_segment)
        self.active_path.current_step = 0
        self.active_path.hold_time_remaining = self.active_path.hold_time
        self.active_path.max_steps = round(self.active_path.total_distance / self.active_path.speed)
        for segment in self.active_path.segments:
            segment.enter_event_triggered = False
            segment.exit_event_triggered = False
        if self.active_path.layer is not None:
            self.character.layer = self.active_path.layer
        self.character.event_handler.handle_event(self.character.event_handler.Event.PATH_ACTIVATED, self.active_path)

    def deactivate_path(self, path: Path) -> None:
        """Unsets the current path if the current path is path.

        Args:
            path (Path): the Path to deactivate
        """
        if self.active_path and self.active_path is path:
            self.active_path = None

    def move(self) -> None:
        """Moves the character one step closer to the target position based on an easing function if present, otherwise linearly."""
        # preserve previous coordinate to allow for clearing the location in the terminal
        self.previous_coord = Coord(self.current_coord.column, self.current_coord.row)

        if not self.active_path or not self.active_path.segments:
            return
        self.current_coord = self.active_path.step(self.character.event_handler)
        if self.active_path.current_step == self.active_path.max_steps:
            if self.active_path.hold_time and self.active_path.hold_time_remaining == self.active_path.hold_time:
                self.character.event_handler.handle_event(
                    self.character.event_handler.Event.PATH_HOLDING, self.active_path
                )
                self.active_path.hold_time_remaining -= 1
                return
            elif self.active_path.hold_time_remaining:
                self.active_path.hold_time_remaining -= 1
                return
            if self.active_path.loop and len(self.active_path.segments) > 1:
                looping_path = self.active_path
                self.deactivate_path(self.active_path)
                self.activate_path(looping_path)
            else:
                self.completed_path = self.active_path
                self.deactivate_path(self.active_path)
                self.character.event_handler.handle_event(
                    self.character.event_handler.Event.PATH_COMPLETE, self.completed_path
                )



"""Effect that draws the characters spawning at varying rates from a single point."""

import argparse
import random
from enum import Enum, auto

import terminaltexteffects.utils.argtypes as argtypes
from terminaltexteffects.base_character import EffectCharacter, EventHandler
from terminaltexteffects.utils import graphics, motion
from terminaltexteffects.utils.terminal import Terminal


def add_arguments(subparsers: argparse._SubParsersAction) -> None:
    """Adds arguments to the subparser.

    Args:
        subparser (argparse._SubParsersAction): subparser to add arguments to
    """
    effect_parser = subparsers.add_parser(
        "spray",
        formatter_class=argtypes.CustomFormatter,
        help="Draws the characters spawning at varying rates from a single point.",
        description="spray | Draws the characters spawning at varying rates from a single point.",
        epilog=f"""{argtypes.EASING_EPILOG}
        
Example: terminaltexteffects spray -a 0.01 --spray-position center""",
    )
    effect_parser.set_defaults(effect_class=SprayEffect)
    effect_parser.add_argument(
        "-a",
        "--animation-rate",
        type=argtypes.valid_animationrate,
        default=0.01,
        help="Minimum time, in seconds, between animation steps. This value does not normally need to be modified. Use this to increase the playback speed of all aspects of the effect. This will have no impact beyond a certain lower threshold due to the processing speed of your device.",
    )
    effect_parser.add_argument(
        "--spray-colors",
        type=argtypes.valid_color,
        nargs="*",
        default=0,
        metavar="(XTerm [0-255] OR RGB Hex [000000-ffffff])",
        help="List of colors for the character spray. Colors are randomly chosen from the list.",
    )
    effect_parser.add_argument(
        "--final-color",
        type=argtypes.valid_color,
        default="ffffff",
        metavar="(XTerm [0-255] OR RGB Hex [000000-ffffff])",
        help="Color for the final character.",
    )
    effect_parser.add_argument(
        "--spray-position",
        default="e",
        choices=["n", "ne", "e", "se", "s", "sw", "w", "nw", "center"],
        help="Position for the spray origin.",
    )
    effect_parser.add_argument(
        "--spray-volume",
        default=5,
        type=argtypes.positive_int,
        metavar="(int > 0)",
        help="Maximum number of characters to spray at a time.",
    )
    effect_parser.add_argument(
        "--movement-speed",
        type=argtypes.valid_speed,
        default=0.7,
        metavar="(float > 0)",
        help="Movement speed of the characters. Note: Speed effects the number of steps in the easing function. Adjust speed and animation rate separately to fine tune the effect.",
    )
    effect_parser.add_argument(
        "--easing",
        default="OUT_EXPO",
        type=argtypes.valid_ease,
        help="Easing function to use for character movement.",
    )


class SprayPosition(Enum):
    """Position for the spray origin."""

    N = auto()
    NE = auto()
    E = auto()
    SE = auto()
    S = auto()
    SW = auto()
    W = auto()
    NW = auto()
    CENTER = auto()


class SprayEffect:
    """Effect that draws the characters spawning at varying rates from a single point."""

    def __init__(
        self,
        terminal: Terminal,
        args: argparse.Namespace,
    ):
        """Effect that draws the characters spawning at varying rates from a single point.

        Args:
            terminal (Terminal): terminal to use for the effect
            args (argparse.Namespace): arguments from argparse
        """
        self.terminal = terminal
        self.args = args
        self.pending_chars: list[EffectCharacter] = []
        self.animating_chars: list[EffectCharacter] = []
        self.spray_position = {
            "n": SprayPosition.N,
            "ne": SprayPosition.NE,
            "e": SprayPosition.E,
            "se": SprayPosition.SE,
            "s": SprayPosition.S,
            "sw": SprayPosition.SW,
            "w": SprayPosition.W,
            "nw": SprayPosition.NW,
            "center": SprayPosition.CENTER,
        }.get(args.spray_position, SprayPosition.E)
        self.spray_colors = args.spray_colors
        self.final_color = args.final_color

    def prepare_data(self) -> None:
        """Prepares the data for the effect by starting all of the characters from a point based on SparklerPosition."""
        spray_origin_map = {
            SprayPosition.CENTER: (self.terminal.output_area.center),
            SprayPosition.N: motion.Coord(self.terminal.output_area.right // 2, self.terminal.output_area.top),
            SprayPosition.NW: motion.Coord(self.terminal.output_area.left, self.terminal.output_area.top),
            SprayPosition.W: motion.Coord(self.terminal.output_area.left, self.terminal.output_area.top // 2),
            SprayPosition.SW: motion.Coord(self.terminal.output_area.left, self.terminal.output_area.bottom),
            SprayPosition.S: motion.Coord(self.terminal.output_area.right // 2, self.terminal.output_area.bottom),
            SprayPosition.SE: motion.Coord(self.terminal.output_area.right - 1, self.terminal.output_area.bottom),
            SprayPosition.E: motion.Coord(self.terminal.output_area.right - 1, self.terminal.output_area.top // 2),
            SprayPosition.NE: motion.Coord(self.terminal.output_area.right - 1, self.terminal.output_area.top),
        }

        for character in self.terminal.characters:
            character.motion.set_coordinate(spray_origin_map[self.spray_position])
            input_coord_path = character.motion.new_path(
                "input_coord", speed=self.args.movement_speed, ease=self.args.easing
            )
            input_coord_wpt = input_coord_path.new_waypoint("input_coord", character.input_coord)
            character.event_handler.register_event(
                EventHandler.Event.PATH_ACTIVATED, input_coord_path, EventHandler.Action.SET_LAYER, 1
            )
            character.event_handler.register_event(
                EventHandler.Event.PATH_COMPLETE, input_coord_path, EventHandler.Action.SET_LAYER, 0
            )
            if self.spray_colors:
                droplet_scn = character.animation.new_scene("droplet")
                spray_color = random.choice(self.spray_colors)
                spray_gradient = graphics.Gradient([spray_color, self.final_color], 7)
                for color in spray_gradient:
                    droplet_scn.add_frame(character.input_symbol, 40, color=color)
                character.animation.activate_scene(droplet_scn)
            character.motion.activate_path(input_coord_path)
            self.pending_chars.append(character)
        random.shuffle(self.pending_chars)

    def run(self) -> None:
        """Runs the effect."""
        self.prepare_data()
        while self.pending_chars or self.animating_chars:
            if self.pending_chars:
                for _ in range(random.randint(1, self.args.spray_volume)):
                    if self.pending_chars:
                        next_character = self.pending_chars.pop()
                        next_character.is_active = True
                        self.animating_chars.append(next_character)

            self.animate_chars()
            self.terminal.print()
            self.animating_chars = [
                animating_char
                for animating_char in self.animating_chars
                if not animating_char.animation.active_scene_is_complete()
                or not animating_char.motion.movement_is_complete()
            ]

    def animate_chars(self) -> None:
        for animating_char in self.animating_chars:
            animating_char.animation.step_animation()
            animating_char.motion.move()



"""Effect that pours the characters into position from the top, bottom, left, or right."""

import argparse
from enum import Enum, auto

import terminaltexteffects.utils.argtypes as argtypes
from terminaltexteffects.base_character import EffectCharacter
from terminaltexteffects.utils.terminal import Terminal
from terminaltexteffects.utils import motion


def add_arguments(subparsers: argparse._SubParsersAction) -> None:
    """Adds arguments to the subparser.

    Args:
        subparser (argparse._SubParsersAction): subparser to add arguments to
    """
    effect_parser = subparsers.add_parser(
        "pour",
        formatter_class=argtypes.CustomFormatter,
        help="Pours the characters into position from the given direction.",
        description="pour | Pours the characters into position from the given direction.",
        epilog=f"""{argtypes.EASING_EPILOG}
        
Example: terminaltexteffects pour -a 0.004 --pour-direction down""",
    )
    effect_parser.set_defaults(effect_class=PourEffect)
    effect_parser.add_argument(
        "-a",
        "--animation-rate",
        type=argtypes.valid_animationrate,
        default=0.004,
        help="Minimum time, in seconds, between animation steps. This value does not normally need to be modified. Use this to increase the playback speed of all aspects of the effect. This will have no impact beyond a certain lower threshold due to the processing speed of your device.",
    )
    effect_parser.add_argument(
        "--pour-direction",
        default="down",
        choices=["up", "down", "left", "right"],
        help="Direction the text will pour.",
    )
    effect_parser.add_argument(
        "--movement-speed",
        type=argtypes.valid_speed,
        default=0.2,
        metavar="(float > 0)",
        help="Movement speed of the characters. Note: Speed effects the number of steps in the easing function. Adjust speed and animation rate separately to fine tune the effect.",
    )
    effect_parser.add_argument(
        "--easing",
        default="IN_QUAD",
        type=argtypes.valid_ease,
        help="Easing function to use for character movement.",
    )


class PourDirection(Enum):
    UP = auto()
    DOWN = auto()
    LEFT = auto()
    RIGHT = auto()


class PourEffect:
    """Effect that pours the characters into position from the top, bottom, left, or right."""

    def __init__(self, terminal: Terminal, args: argparse.Namespace):
        self.terminal = terminal
        self.args = args
        self.pending_chars: list[EffectCharacter] = []
        self.animating_chars: list[EffectCharacter] = []
        self.pour_direction = {
            "down": PourDirection.DOWN,
            "up": PourDirection.UP,
            "left": PourDirection.LEFT,
            "right": PourDirection.RIGHT,
        }.get(args.pour_direction, PourDirection.DOWN)

    def prepare_data(self) -> None:
        """Prepares the data for the effect by sorting the characters by the pour direction."""
        sort_map = {
            PourDirection.DOWN: lambda character: character.input_coord.row,
            PourDirection.UP: lambda character: -character.input_coord.row,
            PourDirection.LEFT: lambda character: character.input_coord.column,
            PourDirection.RIGHT: lambda character: -character.input_coord.column,
        }
        self.terminal.characters.sort(key=sort_map[self.pour_direction])
        for character in self.terminal.characters:
            character.is_active = False
            if self.pour_direction == PourDirection.DOWN:
                character.motion.set_coordinate(
                    motion.Coord(character.input_coord.column, self.terminal.output_area.top)
                )
            elif self.pour_direction == PourDirection.UP:
                character.motion.set_coordinate(
                    motion.Coord(character.input_coord.column, self.terminal.output_area.bottom)
                )
            elif self.pour_direction == PourDirection.LEFT:
                character.motion.set_coordinate(
                    motion.Coord(self.terminal.output_area.right, character.input_coord.row)
                )
            elif self.pour_direction == PourDirection.RIGHT:
                character.motion.set_coordinate(motion.Coord(self.terminal.output_area.left, character.input_coord.row))
            input_coord_path = character.motion.new_path(
                "input_coord",
                speed=self.args.movement_speed,
                ease=self.args.easing,
            )
            input_coord_wpt = input_coord_path.new_waypoint("input_coord", character.input_coord)
            character.motion.activate_path(input_coord_path)
            self.pending_chars.append(character)

    def run(self) -> None:
        """Runs the effect."""
        self.prepare_data()
        self.terminal.print()
        while self.pending_chars or self.animating_chars:
            if self.pending_chars:
                next_character = self.pending_chars.pop(0)
                next_character.is_active = True
                self.animating_chars.append(next_character)
            self.animate_chars()
            self.animating_chars = [
                animating_char
                for animating_char in self.animating_chars
                if not animating_char.motion.movement_is_complete()
            ]
            self.terminal.print()

    def animate_chars(self) -> None:
        """Animates the sliding characters."""
        for animating_char in self.animating_chars:
            animating_char.motion.move()



import argparse

from terminaltexteffects.utils import argtypes, motion
from terminaltexteffects.base_character import EffectCharacter
from terminaltexteffects.utils.terminal import Terminal


def add_arguments(subparsers: argparse._SubParsersAction) -> None:
    """Adds arguments to the subparser.

    Args:
        subparser (argparse._SubParsersAction): subparser to add arguments to
    """
    effect_parser = subparsers.add_parser(
        "verticalslice",
        formatter_class=argtypes.CustomFormatter,
        help="Slices the input in half vertically and slides it into place from opposite directions.",
        description="verticalslice | Slices the input in half vertically and slides it into place from opposite directions.",
        epilog=f"""{argtypes.EASING_EPILOG}
        
Example: terminaltexteffects verticalslice -a 0.02""",
    )
    effect_parser.set_defaults(effect_class=VerticalSlice)
    effect_parser.add_argument(
        "-a",
        "--animation-rate",
        type=argtypes.valid_animationrate,
        default=0.02,
        help="Minimum time, in seconds, between animation steps. This value does not normally need to be modified. Use this to increase the playback speed of all aspects of the effect. This will have no impact beyond a certain lower threshold due to the processing speed of your device.",
    )
    effect_parser.add_argument(
        "--movement-speed",
        type=argtypes.valid_speed,
        default=0.5,
        metavar="(float > 0)",
        help="Movement speed of the characters. Note: Speed effects the number of steps in the easing function. Adjust speed and animation rate separately to fine tune the effect.",
    )
    effect_parser.add_argument(
        "--easing",
        default="IN_OUT_EXPO",
        type=argtypes.valid_ease,
        help="Easing function to use for character movement.",
    )


class VerticalSlice:
    """Effect that slices the input in half vertically and slides it into place from opposite directions."""

    def __init__(self, terminal: Terminal, args: argparse.Namespace):
        self.terminal = terminal
        self.args = args
        self.pending_chars: list[EffectCharacter] = []
        self.animating_chars: list[EffectCharacter] = []

    def prepare_data(self) -> None:
        """Prepares the data for the effect by setting the left half to start at the top and the
        right half to start at the bottom, and creating rows consisting off halves from opposite
        input rows."""

        self.rows = list(self.terminal.input_by_row().values())
        lengths = [max([c.input_coord.column for c in row]) for row in self.rows]
        mid_point = sum(lengths) // len(lengths) // 2
        self.new_rows = []
        for row_index, row in enumerate(self.rows):
            new_row = []
            left_half = [character for character in row if character.input_coord.column <= mid_point]
            for character in left_half:
                character.motion.set_coordinate(
                    motion.Coord(character.input_coord.column, self.terminal.output_area.top)
                )
                input_coord_path = character.motion.new_path(
                    "input_coord", speed=self.args.movement_speed, ease=self.args.easing
                )
                input_coord_wpt = input_coord_path.new_waypoint("input_coord", character.input_coord)
                character.motion.activate_path(input_coord_path)
            opposite_row = self.rows[-(row_index + 1)]
            right_half = [c for c in opposite_row if c.input_coord.column > mid_point]
            for character in right_half:
                character.motion.set_coordinate(
                    motion.Coord(character.input_coord.column, self.terminal.output_area.bottom)
                )
                input_coord_path = character.motion.new_path(
                    "input_coord", speed=self.args.movement_speed, ease=self.args.easing
                )
                input_coord_wpt = input_coord_path.new_waypoint("input_coord", character.input_coord)
                character.motion.activate_path(input_coord_path)
            new_row.extend(left_half)
            new_row.extend(right_half)
            self.new_rows.append(new_row)

    def run(self) -> None:
        """Runs the effect."""
        self.prepare_data()
        while self.new_rows or self.animating_chars:
            if self.new_rows:
                next_row = self.new_rows.pop(0)
                for character in next_row:
                    character.is_active = True
                self.animating_chars.extend(next_row)
            self.animate_chars()
            self.terminal.print()
            # remove completed chars from animating chars
            self.animating_chars = [
                animating_char
                for animating_char in self.animating_chars
                if not animating_char.motion.movement_is_complete()
            ]

    def animate_chars(self) -> None:
        """Animates the characters by calling the move method and printing the characters to the terminal."""
        for animating_char in self.animating_chars:
            animating_char.motion.move()



import argparse
import random

import terminaltexteffects.utils.argtypes as argtypes
from terminaltexteffects.base_character import EffectCharacter, EventHandler
from terminaltexteffects.utils.terminal import Terminal
from terminaltexteffects.utils import graphics, motion, argtypes, easing


def add_arguments(subparsers: argparse._SubParsersAction) -> None:
    """Adds arguments to the subparser.

    Args:
        subparser (argparse._SubParsersAction): subparser to add arguments to
    """
    effect_parser = subparsers.add_parser(
        "crumble",
        formatter_class=argtypes.CustomFormatter,
        help="Characters lose color and crumble into dust, vacuumed up, and reformed.",
        description="Characters lose color and crumble into dust, vacuumed up, and reformed.",
        epilog=f"""{argtypes.EASING_EPILOG}

Example: terminaltexteffects --effect crumble --animation-rate 0.01 --initial-color 0088bb --dust-colors dadad1 766b69 848789 747a8a --final-color 0088bb""",
    )
    effect_parser.set_defaults(effect_class=CrumbleEffect)
    effect_parser.add_argument(
        "-a",
        "--animation-rate",
        type=argtypes.valid_animationrate,
        default=0.01,
        help="Minimum time, in seconds, between animation steps. This value does not normally need to be modified. Use this to increase the playback speed of all aspects of the effect. This will have no impact beyond a certain lower threshold due to the processing speed of your device.",
    )
    effect_parser.add_argument(
        "--initial-color",
        type=argtypes.valid_color,
        default="0088bb",
        metavar="(XTerm [0-255] OR RGB Hex [000000-ffffff])",
        help="Color for the characters at the start of the effect.",
    )
    effect_parser.add_argument(
        "--dust-colors",
        type=argtypes.valid_color,
        nargs="*",
        default=["dadad1", "766b69", "848789", "747a8a"],
        metavar="(XTerm [0-255] OR RGB Hex [000000-ffffff])",
        help="Space separated, unquoted, list of colors for the dust particles.",
    )
    effect_parser.add_argument(
        "--final-color",
        type=argtypes.valid_color,
        default="0eeaff",
        metavar="(XTerm [0-255] OR RGB Hex [000000-ffffff])",
        help="Color for the final character.",
    )


class CrumbleEffect:
    """Characters crumble into dust before being vacuumed up and rebuilt."""

    def __init__(self, terminal: Terminal, args: argparse.Namespace):
        self.terminal = terminal
        self.args = args
        self.pending_chars: list[EffectCharacter] = []
        self.animating_chars: list[EffectCharacter] = []

    def prepare_data(self) -> None:
        """Prepares the data for the effect by registering events and building scenes/waypoints."""
        strengthen_flash_gradient = graphics.Gradient([self.args.final_color, "ffffff"], 6)
        strengthen_gradient = graphics.Gradient(["ffffff", self.args.final_color], 9)

        for character in self.terminal.characters:
            dust_color = random.choice(self.args.dust_colors)
            weaken_gradient = graphics.Gradient([self.args.initial_color, dust_color], 9)
            character.is_active = True
            # set up initial and falling stage
            initial_scn = character.animation.new_scene("initial")
            initial_scn.add_frame(character.input_symbol, 1, color=self.args.initial_color)
            character.animation.activate_scene(initial_scn)
            fall_path = character.motion.new_path(
                "fall",
                speed=0.2,
                ease=easing.out_bounce,
            )
            fall_wpt = fall_path.new_waypoint(
                "fall", motion.Coord(character.input_coord.column, self.terminal.output_area.bottom)
            )
            weaken_scn = character.animation.new_scene("weaken")
            for color_step in weaken_gradient:
                weaken_scn.add_frame(character.input_symbol, 6, color=color_step)

            # set up vacuuming stage
            # bottom_vac_path = character.motion.new_path("bottom_vac", speed=0.3, ease=easing.in_quint)
            # bottom_vac_wpt = bottom_vac_path.new_waypoint(
            #     "bottom_vac",
            #     motion.Coord(
            #         self.terminal.output_area.center_column,
            #         round(self.terminal.output_area.top * random.uniform(0.1, 0.3)),
            #     ),
            #     bezier_control=motion.Coord(self.terminal.output_area.center_column, 1),
            # )
            # top_vac_path = character.motion.new_path("top_vac", speed=0.3)
            # top_vac_wpt = top_vac_path.new_waypoint(
            #     "top_vac",
            #     motion.Coord(
            #         self.terminal.output_area.center_column,
            #         round(self.terminal.output_area.top * random.uniform(0.7, 0.9)),
            #     ),
            # )
            top_path = character.motion.new_path("top", speed=0.3, ease=easing.out_quint)
            top_wpt = top_path.new_waypoint(
                "top",
                motion.Coord(character.input_coord.column, self.terminal.output_area.top),
                bezier_control=motion.Coord(
                    self.terminal.output_area.center_column, self.terminal.output_area.center_row
                ),
            )
            # set up reset stage
            input_path = character.motion.new_path("input", speed=0.3)
            input_wpt = input_path.new_waypoint("input", character.input_coord)
            strengthen_flash_scn = character.animation.new_scene("strengthen_flash")
            for color_step in strengthen_flash_gradient:
                strengthen_flash_scn.add_frame(character.input_symbol, 4, color=color_step)
            strengthen_scn = character.animation.new_scene("strengthen")
            for color_step in strengthen_gradient:
                strengthen_scn.add_frame(character.input_symbol, 6, color=color_step)
            dust_scn = character.animation.new_scene("dust", sync=graphics.SyncMetric.DISTANCE)
            for _ in range(5):
                dust_scn.add_frame(random.choice(["*", ".", ","]), 1, color=dust_color)

            character.event_handler.register_event(
                EventHandler.Event.SCENE_COMPLETE, weaken_scn, EventHandler.Action.ACTIVATE_PATH, fall_path
            )
            character.event_handler.register_event(
                EventHandler.Event.SCENE_COMPLETE, weaken_scn, EventHandler.Action.SET_LAYER, 1
            )
            character.event_handler.register_event(
                EventHandler.Event.SCENE_COMPLETE, weaken_scn, EventHandler.Action.ACTIVATE_SCENE, dust_scn
            )
            # character.event_handler.register_event(
            #     EventHandler.Event.PATH_COMPLETE, bottom_vac_path, EventHandler.Action.ACTIVATE_PATH, top_vac_path
            # )
            # character.event_handler.register_event(
            #     EventHandler.Event.PATH_COMPLETE, top_vac_path, EventHandler.Action.ACTIVATE_PATH, top_path
            # )
            character.event_handler.register_event(
                EventHandler.Event.PATH_COMPLETE,
                input_path,
                EventHandler.Action.ACTIVATE_SCENE,
                strengthen_flash_scn,
            )
            character.event_handler.register_event(
                EventHandler.Event.SCENE_COMPLETE,
                strengthen_flash_scn,
                EventHandler.Action.ACTIVATE_SCENE,
                strengthen_scn,
            )
            self.pending_chars.append(character)

    def run(self) -> None:
        """Runs the effect."""
        # Prepare data for the effect
        self.prepare_data()
        random.shuffle(self.pending_chars)
        fall_delay = 20
        max_fall_delay = 20
        min_fall_delay = 15
        vacuumed = False
        reset = False
        fall_group_maxsize = 1
        stage = "falling"
        unvacuumed_chars = list(self.terminal.characters)
        random.shuffle(unvacuumed_chars)
        self.terminal.print()
        while stage != "complete":
            if stage == "falling":
                if self.pending_chars:
                    if fall_delay == 0:
                        # Determine the size of the next group of falling characters
                        fall_group_size = random.randint(1, fall_group_maxsize)
                        # Add the next group of falling characters to the animating characters list
                        for _ in range(fall_group_size):
                            if self.pending_chars:
                                next_char = self.pending_chars.pop(0)
                                next_char.animation.activate_scene(next_char.animation.scenes["weaken"])
                                self.animating_chars.append(next_char)
                        # Reset the fall delay and adjust the fall group size and delay range
                        fall_delay = random.randint(min_fall_delay, max_fall_delay)
                        if random.randint(1, 10) > 4:  # 60% chance to modify the fall delay and group size
                            fall_group_maxsize += 1
                            min_fall_delay = max(0, min_fall_delay - 1)
                            max_fall_delay = max(0, max_fall_delay - 1)
                    else:
                        fall_delay -= 1
                if not self.pending_chars and not self.animating_chars:
                    stage = "vacuuming"
            elif stage == "vacuuming":
                if unvacuumed_chars:
                    for _ in range(random.randint(3, 10)):
                        if unvacuumed_chars:
                            next_char = unvacuumed_chars.pop(0)
                            next_char.motion.activate_path(next_char.motion.paths["top"])
                            self.animating_chars.append(next_char)
                if not self.animating_chars:
                    stage = "resetting"

            elif stage == "resetting":
                if not reset:
                    for character in self.terminal.characters:
                        character.motion.activate_path(character.motion.paths["input"])
                        self.animating_chars.append(character)
                    reset = True
                if not self.animating_chars:
                    stage = "complete"

            self.terminal.print()
            self.animate_chars()
            self.animating_chars = [
                animating_char
                for animating_char in self.animating_chars
                if not animating_char.animation.active_scene_is_complete()
                or not animating_char.motion.movement_is_complete()
            ]

    def animate_chars(self) -> None:
        """Animates the characters by calling the move method and step animation. Move characters prior to stepping animation
        to ensure waypoint synced animations have the latest waypoint progress information."""
        for animating_char in self.animating_chars:
            animating_char.motion.move()
            animating_char.animation.step_animation()



import argparse
from enum import Enum, auto

from terminaltexteffects.utils import argtypes, motion
from terminaltexteffects.base_character import EffectCharacter
from terminaltexteffects.utils.terminal import Terminal


def add_arguments(subparsers: argparse._SubParsersAction) -> None:
    """Adds arguments to the subparser.

    Args:
        subparser (argparse._SubParsersAction): subparser to add arguments to
    """
    effect_parser = subparsers.add_parser(
        "rowslide",
        formatter_class=argtypes.CustomFormatter,
        help="Slides each row into place.",
        description="rowslide | Slides each row into place.",
        epilog=f"""{argtypes.EASING_EPILOG}
        
Example: terminaltexteffects rowslide -a 0.003 --slide-direction left""",
    )
    effect_parser.set_defaults(effect_class=RowSlide)
    effect_parser.add_argument(
        "-a",
        "--animation-rate",
        type=argtypes.valid_animationrate,
        default=0.01,
        help="Minimum time, in seconds, between animation steps. This value does not normally need to be modified. Use this to increase the playback speed of all aspects of the effect. This will have no impact beyond a certain lower threshold due to the processing speed of your device.",
    )
    effect_parser.add_argument(
        "--row-gap",
        default=5,
        type=argtypes.valid_gap,
        help="Number of characters to wait before adding a new row. Min 1.",
    )
    effect_parser.add_argument(
        "--slide-direction",
        default="left",
        choices=["left", "right"],
        help="Direction the text will slide.",
    )
    effect_parser.add_argument(
        "--movement-speed",
        type=argtypes.valid_speed,
        default=0.8,
        metavar="(float > 0)",
        help="Movement speed of the characters. Note: Speed effects the number of steps in the easing function. Adjust speed and animation rate separately to fine tune the effect.",
    )
    effect_parser.add_argument(
        "--easing",
        default="IN_OUT_QUAD",
        type=argtypes.valid_ease,
        help="Easing function to use for row movement.",
    )


class SlideDirection(Enum):
    LEFT = auto()
    RIGHT = auto()


class RowSlide:
    """Effect that slides each row into place."""

    def __init__(self, terminal: Terminal, args: argparse.Namespace):
        """Effect that slides each row into place.

        Args:
            terminal (Terminal): terminal to use for the effect
            args (argparse.Namespace): arguments from argparse
        """
        self.terminal = terminal
        self.args = args
        self.pending_chars: list[EffectCharacter] = []
        self.animating_chars: list[EffectCharacter] = []
        self.row_gap: int = args.row_gap
        if args.slide_direction == "left":
            self.slide_direction = SlideDirection.LEFT
        else:
            self.slide_direction = SlideDirection.RIGHT

    def prepare_data(self) -> None:
        """Prepares the data for the effect by grouping the characters by row and setting the starting
        coordinate."""

        self.rows = self.terminal.input_by_row()
        for row in self.rows.values():
            for character in row:
                if self.slide_direction == SlideDirection.LEFT:
                    character.motion.set_coordinate(
                        motion.Coord(self.terminal.output_area.right, character.input_coord.row)
                    )
                else:
                    character.motion.set_coordinate(motion.Coord(0, character.input_coord.row))
                input_coord_path = character.motion.new_path(
                    "input_coord", speed=self.args.movement_speed, ease=self.args.easing
                )
                input_coord_path.new_waypoint("input_coord", character.input_coord)
                character.motion.activate_path(input_coord_path)

    def get_next_row(self) -> list[EffectCharacter]:
        """Gets the next row of characters to animate.

        Returns:
            list[effect_char.EffectCharacter]: The next row of characters to animate.
        """
        next_row = self.rows.pop(min(self.rows.keys()))
        return next_row

    def run(self) -> None:
        """Runs the effect."""
        self.prepare_data()
        active_rows: list[list[EffectCharacter]] = []
        active_rows.append(self.get_next_row())
        row_delay_countdown = self.row_gap
        while active_rows or self.animating_chars:
            if (row_delay_countdown == 0 and self.rows) or (not active_rows and self.rows):
                active_rows.append(self.get_next_row())
                row_delay_countdown = self.row_gap
            else:
                if self.rows:
                    row_delay_countdown -= 1
            for row in active_rows:
                if row:
                    if self.slide_direction == SlideDirection.LEFT:
                        next_character = row.pop(0)
                        next_character.is_active = True
                        self.animating_chars.append(next_character)
                    else:
                        next_character = row.pop(-1)
                        next_character.is_active = True
                        self.animating_chars.append(next_character)
            self.animate_chars()
            self.terminal.print()
            self.animating_chars = [
                animating_char
                for animating_char in self.animating_chars
                if not animating_char.motion.movement_is_complete()
            ]
            active_rows = [row for row in active_rows if row]

    def animate_chars(self) -> None:
        """Animates the characters by calling the move method."""
        for animating_char in self.animating_chars:
            animating_char.motion.move()



import argparse
import random

import terminaltexteffects.utils.argtypes as argtypes
from terminaltexteffects.base_character import EffectCharacter, EventHandler
from terminaltexteffects.utils.terminal import Terminal
from terminaltexteffects.utils import graphics, argtypes


def add_arguments(subparsers: argparse._SubParsersAction) -> None:
    """Adds arguments to the subparser.

    Args:
        subparser (argparse._SubParsersAction): subparser to add arguments to
    """
    effect_parser = subparsers.add_parser(
        "errorcorrect",
        formatter_class=argtypes.CustomFormatter,
        help="Some characters start in the wrong position and are corrected in sequence.",
        description="Some characters start in the wrong position and are corrected in sequence.",
        epilog=f"""{argtypes.EASING_EPILOG}

Example: terminaltexteffects errorcorrect -a 0.01 --error-pairs 12 --swap-delay 70 --error-color e74c3c --correct-color 45bf55 
         --final-color ffffff --movement-speed 0.5""",
    )
    effect_parser.set_defaults(effect_class=ErrorCorrectEffect)
    effect_parser.add_argument(
        "-a",
        "--animation-rate",
        type=argtypes.valid_animationrate,
        default=0.01,
        help="Minimum time, in seconds, between animation steps. This value does not normally need to be modified. Use this to increase the playback speed of all aspects of the effect. This will have no impact beyond a certain lower threshold due to the processing speed of your device.",
    )
    effect_parser.add_argument(
        "--error-pairs",
        type=argtypes.positive_int,
        default=12,
        metavar="(int > 0)",
        help="Number of character pairs to swap.",
    )
    effect_parser.add_argument(
        "--swap-delay",
        type=argtypes.positive_int,
        default=70,
        metavar="(int > 0)",
        help="Number of animation steps between swaps.",
    )
    effect_parser.add_argument(
        "--error-color",
        type=argtypes.valid_color,
        default="e74c3c",
        metavar="(XTerm [0-255] OR RGB Hex [000000-ffffff])",
        help="Color for the characters that are in the wrong position.",
    )
    effect_parser.add_argument(
        "--correct-color",
        type=argtypes.valid_color,
        default="45bf55",
        metavar="(XTerm [0-255] OR RGB Hex [000000-ffffff])",
        help="Color for the characters once corrected, this is a gradient from error-color and fades to final-color.",
    )
    effect_parser.add_argument(
        "--final-color",
        type=argtypes.valid_color,
        default="ffffff",
        metavar="(XTerm [0-255] OR RGB Hex [000000-ffffff])",
        help="Color for the characters in their correct position.",
    )
    effect_parser.add_argument(
        "--movement-speed",
        type=argtypes.valid_speed,
        default=0.5,
        metavar="(float > 0)",
        help="Speed of the characters while moving to the correct position. Note: Speed effects the number of steps in the easing function. Adjust speed and animation rate separately to fine tune the effect.",
    )


class ErrorCorrectEffect:
    """Effect that swaps characters from an incorrect initial position to the correct position."""

    def __init__(self, terminal: Terminal, args: argparse.Namespace):
        self.terminal = terminal
        self.args = args
        self.pending_chars: list[EffectCharacter] = []
        self.animating_chars: list[EffectCharacter] = []
        self.swapped: list[tuple[EffectCharacter, EffectCharacter]] = []

    def prepare_data(self) -> None:
        """Prepares the data for the effect by swapping positions and generating animations and waypoints."""
        for character in self.terminal.characters:
            spawn_scene = character.animation.new_scene("spawn")
            spawn_scene.add_frame(character.input_symbol, 1, color=self.args.final_color)
            character.animation.activate_scene(spawn_scene)
            character.is_active = True
        all_characters: list[EffectCharacter] = list(self.terminal.characters)
        correcting_gradient = graphics.Gradient([self.args.error_color, self.args.correct_color], 10)
        final_gradient = graphics.Gradient([self.args.correct_color, self.args.final_color], 10)
        block_symbol = "▓"
        block_wipe_start = ("▁", "▂", "▃", "▄", "▅", "▆", "▇", "█")
        block_wipe_end = ("▇", "▆", "▅", "▄", "▃", "▂", "▁")
        for _ in range(self.args.error_pairs):
            if len(all_characters) < 2:
                break
            char1 = all_characters.pop(random.randrange(len(all_characters)))
            char2 = all_characters.pop(random.randrange(len(all_characters)))
            char1.motion.set_coordinate(char2.input_coord)
            char1_input_coord_path = char1.motion.new_path("input_coord", speed=self.args.movement_speed)
            char1_input_coord_waypoint = char1_input_coord_path.new_waypoint("input_coord", char1.input_coord)
            char2.motion.set_coordinate(char1.input_coord)
            char2_input_coord_path = char2.motion.new_path("input_coord", speed=self.args.movement_speed)
            char2_input_coord_waypoint = char2_input_coord_path.new_waypoint("input_coord", char2.input_coord)
            self.swapped.append((char1, char2))
            for character in (char1, char2):
                first_block_wipe = character.animation.new_scene("first_block_wipe")
                last_block_wipe = character.animation.new_scene("last_block_wipe")
                for block in block_wipe_start:
                    first_block_wipe.add_frame(block, 3, color=self.args.error_color)
                for block in block_wipe_end:
                    last_block_wipe.add_frame(block, 3, color=self.args.correct_color)
                initial_scene = character.animation.new_scene("initial")
                initial_scene.add_frame(character.input_symbol, 1, color=self.args.error_color)
                character.animation.activate_scene(initial_scene)
                error_scene = character.animation.new_scene("error")
                for _ in range(10):
                    error_scene.add_frame(block_symbol, 3, color=self.args.error_color)
                    error_scene.add_frame(character.input_symbol, 3, color="ffffff")
                correcting_scene = character.animation.new_scene("correcting", sync=graphics.SyncMetric.DISTANCE)
                for step in correcting_gradient:
                    correcting_scene.add_frame("█", 3, color=step)
                final_scene = character.animation.new_scene("final")
                for step in final_gradient:
                    final_scene.add_frame(character.input_symbol, 3, color=step)
                character.event_handler.register_event(
                    EventHandler.Event.SCENE_COMPLETE,
                    error_scene,
                    EventHandler.Action.ACTIVATE_SCENE,
                    first_block_wipe,
                )
                character.event_handler.register_event(
                    EventHandler.Event.SCENE_COMPLETE,
                    first_block_wipe,
                    EventHandler.Action.ACTIVATE_SCENE,
                    correcting_scene,
                )
                character.event_handler.register_event(
                    EventHandler.Event.SCENE_COMPLETE,
                    first_block_wipe,
                    EventHandler.Action.ACTIVATE_PATH,
                    character.motion.paths["input_coord"],
                )
                character.event_handler.register_event(
                    EventHandler.Event.PATH_ACTIVATED,
                    character.motion.paths["input_coord"],
                    EventHandler.Action.SET_LAYER,
                    1,
                )
                character.event_handler.register_event(
                    EventHandler.Event.PATH_COMPLETE,
                    character.motion.paths["input_coord"],
                    EventHandler.Action.SET_LAYER,
                    0,
                )

                character.event_handler.register_event(
                    EventHandler.Event.PATH_COMPLETE,
                    character.motion.paths["input_coord"],
                    EventHandler.Action.ACTIVATE_SCENE,
                    last_block_wipe,
                )
                character.event_handler.register_event(
                    EventHandler.Event.SCENE_COMPLETE,
                    last_block_wipe,
                    EventHandler.Action.ACTIVATE_SCENE,
                    final_scene,
                )

    def run(self) -> None:
        """Runs the effect."""
        self.prepare_data()
        while self.swapped or self.animating_chars:
            swap_delay = self.args.swap_delay
            if self.swapped:
                next_pair = self.swapped.pop(0)
                for char in next_pair:
                    char.animation.activate_scene(char.animation.scenes["error"])
                    self.animating_chars.append(char)
            while swap_delay:
                self.terminal.print()
                self.animate_chars()

                self.animating_chars = [
                    animating_char
                    for animating_char in self.animating_chars
                    if not animating_char.animation.active_scene_is_complete()
                    or not animating_char.motion.movement_is_complete()
                ]
                if not self.animating_chars:
                    pass
                swap_delay -= 1
        self.terminal.print()

    def animate_chars(self) -> None:
        """Animates the characters by calling the move method and step animation. Move characters prior to stepping animation
        to ensure waypoint synced animations have the latest waypoint progress information."""
        for animating_char in self.animating_chars:
            animating_char.motion.move()
            animating_char.animation.step_animation()



import argparse
import random

from terminaltexteffects.base_character import EffectCharacter, EventHandler
from terminaltexteffects.utils.terminal import Terminal
from terminaltexteffects.utils import graphics, argtypes, motion, easing


def add_arguments(subparsers: argparse._SubParsersAction) -> None:
    """Adds arguments to the subparser.

    Args:
        subparser (argparse._SubParsersAction): subparser to add arguments to
    """
    effect_parser = subparsers.add_parser(
        "fireworks",
        formatter_class=argtypes.CustomFormatter,
        help="Characters launch and explode like fireworks and fall into place.",
        description="fireworks | Characters explode like fireworks and fall into place.",
        epilog=f"""        
Example: terminaltexteffects fireworks -a 0.01 --firework-colors 88F7E2 44D492 F5EB67 FFA15C FA233E --firework-symbol o --firework-volume 2 --final-color ffffff --launch-delay 60 --explode-distance 10""",
    )
    effect_parser.set_defaults(effect_class=FireworksEffect)
    effect_parser.add_argument(
        "-a",
        "--animation-rate",
        type=argtypes.valid_animationrate,
        default=0.01,
        help="Minimum time, in seconds, between animation steps. This value does not normally need to be modified. Use this to increase the playback speed of all aspects of the effect. This will have no impact beyond a certain lower threshold due to the processing speed of your device.",
    )
    effect_parser.add_argument(
        "--explode-anywhere",
        action="store_true",
        default=False,
        help="If set, fireworks explode anywhere in the output area. Otherwise, fireworks explode above highest settled row of text.",
    )
    effect_parser.add_argument(
        "--firework-colors",
        type=argtypes.valid_color,
        nargs="*",
        default=["88F7E2", "44D492", "F5EB67", "FFA15C", "FA233E"],
        metavar="(XTerm [0-255] OR RGB Hex [000000-ffffff])",
        help="Space separated list of colors from which firework colors will be randomly selected.",
    )
    effect_parser.add_argument(
        "--firework-symbol",
        type=argtypes.valid_symbol,
        default="o",
        metavar="(single character)",
        help="Symbol to use for the firework shell.",
    )
    effect_parser.add_argument(
        "--firework-volume",
        type=argtypes.positive_int,
        default=2,
        metavar="(int > 0)",
        help="Percent of total characters in each firework shell.",
    )
    effect_parser.add_argument(
        "--final-color",
        type=argtypes.valid_color,
        default="ffffff",
        metavar="(XTerm [0-255] OR RGB Hex [000000-ffffff])",
        help="Color for the final character.",
    )
    effect_parser.add_argument(
        "--launch-delay",
        type=argtypes.ge_zero,
        default=60,
        metavar="(int >= 0)",
        help="Number of animation steps to wait between launching each firework shell.",
    )
    effect_parser.add_argument(
        "--explode-distance",
        default=10,
        type=argtypes.positive_int,
        metavar="(int > 0)",
        help="Maximum distance from the firework shell origin to the explode waypoint as a percentage of the total output area width.",
    )


class FireworksEffect:
    """Effect that launches characters up the screen where they explode like fireworks and fall into place."""

    def __init__(self, terminal: Terminal, args: argparse.Namespace):
        self.terminal = terminal
        self.args = args
        self.pending_chars: list[EffectCharacter] = []
        self.animating_chars: list[EffectCharacter] = []
        self.shells: list[list[EffectCharacter]] = []
        if self.args.firework_colors:
            self.firework_colors = self.args.firework_colors
        else:
            self.firework_colors = ["88F7E2", "44D492", "F5EB67", "FFA15C", "FA233E"]
        self.firework_volume = max(1, round((self.args.firework_volume * 0.01) * len(self.terminal.characters)))
        self.explode_distance = min(
            max(1, round(self.terminal.output_area.right * (self.args.explode_distance * 0.01))), 6
        )

    def prepare_waypoints(self) -> None:
        firework_shell: list[EffectCharacter] = []
        for character in self.terminal.characters:
            if len(firework_shell) == self.firework_volume or not firework_shell:
                self.shells.append(firework_shell)
                firework_shell = []
                origin_x = random.randrange(0, self.terminal.output_area.right)
                if not self.args.explode_anywhere:
                    min_row = character.input_coord.row
                else:
                    min_row = self.terminal.output_area.bottom
                origin_y = random.randrange(min_row, self.terminal.output_area.top + 1)
                origin_coord = motion.Coord(origin_x, origin_y)
                explode_waypoint_coords = motion.Motion.find_coords_in_circle(origin_coord, self.explode_distance)
            character.motion.set_coordinate(motion.Coord(origin_x, self.terminal.output_area.bottom))
            apex_path = character.motion.new_path("apex_pth", speed=0.2, ease=easing.out_expo)
            apex_wpt = apex_path.new_waypoint("apex_wpt", origin_coord)
            explode_path = character.motion.new_path("explode_pth", speed=0.15, ease=easing.out_circ)
            explode_wpt = explode_path.new_waypoint("explode_wpt", random.choice(explode_waypoint_coords))

            # TODO: Turn this process into a framework function for making curves when changing direction
            bloom_control_point = motion.Motion.find_coord_at_distance(
                apex_wpt.coord, explode_wpt.coord, self.explode_distance // 2
            )
            bloom_wpt = explode_path.new_waypoint(
                "bloom",
                motion.Coord(bloom_control_point.column, max(1, bloom_control_point.row - 7)),
                bezier_control=bloom_control_point,
            )
            input_path = character.motion.new_path("input_pth", speed=0.3, ease=easing.in_out_quart)
            input_control_point = motion.Coord(bloom_wpt.coord.column, 1)
            input_coord_wpt = input_path.new_waypoint(
                "input_coord", character.input_coord, bezier_control=input_control_point
            )
            character.event_handler.register_event(
                EventHandler.Event.PATH_ACTIVATED, apex_path, EventHandler.Action.SET_LAYER, 2
            )
            character.event_handler.register_event(
                EventHandler.Event.PATH_COMPLETE, explode_path, EventHandler.Action.SET_LAYER, 0
            )
            character.event_handler.register_event(
                EventHandler.Event.PATH_COMPLETE,
                apex_path,
                EventHandler.Action.ACTIVATE_PATH,
                explode_path,
            )
            character.event_handler.register_event(
                EventHandler.Event.PATH_COMPLETE, explode_path, EventHandler.Action.ACTIVATE_PATH, input_path
            )

            character.motion.activate_path(apex_path)

            firework_shell.append(character)
        if firework_shell:
            self.shells.append(firework_shell)

    def prepare_scenes(self) -> None:
        for firework_shell in self.shells:
            shell_color = random.choice(self.firework_colors)
            for character in firework_shell:
                # launch scene
                launch_scn = character.animation.new_scene("launch")
                for color in self.firework_colors:
                    launch_scn.add_frame(self.args.firework_symbol, 2, color=shell_color)
                    launch_scn.add_frame(self.args.firework_symbol, 1, color="FFFFFF")
                    launch_scn.is_looping = True
                # bloom scene
                bloom_scn = character.animation.new_scene("bloom")
                bloom_scn.add_frame(character.input_symbol, 1, color=shell_color)
                # fall scene
                fall_scn = character.animation.new_scene("fall")
                fall_gradient = graphics.Gradient([shell_color, self.args.final_color], 15)
                for color in fall_gradient:
                    fall_scn.add_frame(character.input_symbol, 15, color=color)
                character.animation.activate_scene(launch_scn)
                character.event_handler.register_event(
                    EventHandler.Event.PATH_COMPLETE,
                    character.motion.paths["apex_pth"],
                    EventHandler.Action.ACTIVATE_SCENE,
                    bloom_scn,
                )
                character.event_handler.register_event(
                    EventHandler.Event.PATH_ACTIVATED,
                    character.motion.paths["input_pth"],
                    EventHandler.Action.ACTIVATE_SCENE,
                    fall_scn,
                )

    def prepare_data(self) -> None:
        """Prepares the data for the effect by building the firework shells and scenes."""
        self.prepare_waypoints()
        self.prepare_scenes()

    def run(self) -> None:
        """Runs the effect."""
        self.prepare_data()
        launch_delay = 0
        while self.shells or self.animating_chars:
            if self.shells and launch_delay == 0:
                next_group = self.shells.pop()
                for character in next_group:
                    character.is_active = True
                    self.animating_chars.append(character)
                launch_delay = self.args.launch_delay + 1
            self.terminal.print()
            self.animate_chars()
            launch_delay -= 1

            # remove completed chars from animating chars
            self.animating_chars = [
                animating_char
                for animating_char in self.animating_chars
                if not animating_char.animation.active_scene_is_complete()
                or not animating_char.motion.movement_is_complete()
            ]
        self.terminal.print()

    def animate_chars(self) -> None:
        """Animates the characters by calling the move method and step animation."""
        for animating_char in self.animating_chars:
            animating_char.motion.move()
            animating_char.animation.step_animation()



import argparse
import random
from dataclasses import dataclass

from terminaltexteffects.base_character import EffectCharacter, EventHandler
from terminaltexteffects.utils import graphics, argtypes, motion
from terminaltexteffects.utils.terminal import Terminal


def add_arguments(subparsers: argparse._SubParsersAction) -> None:
    """Adds arguments to the subparser.

    Args:
        subparser (argparse._SubParsersAction): subparser to add arguments to
    """
    effect_parser = subparsers.add_parser(
        "decrypt",
        help="Display a movie style decryption effect.",
        description="decrypt | Movie style decryption effect.",
        epilog="Example: terminaltexteffects decrypt -a 0.003 --cipher-text-color 40 --plain-text-color 208",
    )
    effect_parser.set_defaults(effect_class=DecryptEffect)
    effect_parser.add_argument(
        "-a",
        "--animation-rate",
        type=argtypes.valid_animationrate,
        default=0.003,
        help="Time to sleep between animation steps. Defaults to 0.003 seconds.",
    )
    effect_parser.add_argument(
        "--ciphertext-color",
        type=argtypes.valid_color,
        default=40,
        metavar="(XTerm [0-255] OR RGB Hex [000000-ffffff])",
        help="Color for the ciphertext. Defaults to 40",
    )
    effect_parser.add_argument(
        "--plaintext-color",
        type=argtypes.valid_color,
        default="ffb007",
        metavar="(XTerm [0-255] OR RGB Hex [000000-ffffff])",
        help="Color for the plaintext. Defaults to 208.",
    )


@dataclass
class DecryptChars:
    """Various decimal utf-8 character ranges."""

    keyboard = list(range(33, 127))
    blocks = list(range(9608, 9632))
    box_drawing = list(range(9472, 9599))
    misc = list(range(174, 452))


class DecryptEffect:
    """Effect that shows a movie style text decryption effect."""

    def __init__(self, terminal: Terminal, args: argparse.Namespace):
        self.terminal = terminal
        self.args = args
        self.pending_chars: list[EffectCharacter] = []
        self.animating_chars: list[EffectCharacter] = []
        self.ciphertext_color = args.ciphertext_color
        self.plaintext_color = args.plaintext_color
        self.character_discovered_gradient: graphics.Gradient = graphics.Gradient(["ffffff", self.plaintext_color], 5)
        self.encrypted_symbols: list[str] = []
        self.scenes: dict[str, graphics.Scene] = {}
        self.make_encrypted_symbols()

    def make_encrypted_symbols(self) -> None:
        for n in DecryptChars.keyboard:
            self.encrypted_symbols.append(chr(n))
        for n in DecryptChars.blocks:
            self.encrypted_symbols.append(chr(n))
        for n in DecryptChars.box_drawing:
            self.encrypted_symbols.append(chr(n))
        for n in DecryptChars.misc:
            self.encrypted_symbols.append(chr(n))

    def make_decrypting_animation_scenes(self, character: EffectCharacter) -> None:
        fast_decrypt_scene = character.animation.new_scene("fast_decrypt")
        for _ in range(80):
            symbol = random.choice(self.encrypted_symbols)
            fast_decrypt_scene.add_frame(symbol, 3, color=self.ciphertext_color)
            duration = 3
        slow_decrypt_scene = character.animation.new_scene("slow_decrypt")
        for _ in range(random.randint(1, 15)):  # 1-15 longer duration units
            symbol = random.choice(self.encrypted_symbols)
            if random.randint(0, 100) <= 30:  # 30% chance of extra long duration
                duration = random.randrange(75, 225)  # wide long duration range reduces 'waves' in the animation
            else:
                duration = random.randrange(5, 10)  # shorter duration creates flipping effect
            slow_decrypt_scene.add_frame(symbol, duration, color=self.ciphertext_color)
        discovered_scene = character.animation.new_scene("discovered")
        for color in self.character_discovered_gradient:
            discovered_scene.add_frame(character.input_symbol, 20, color=color)

        discovered_scene.add_frame(character.input_symbol, 1, color=self.plaintext_color)

    def prepare_data_for_type_effect(self) -> None:
        """Prepares the data for the effect by building the animation for each character."""
        for character in self.terminal.characters:
            typing_scene = character.animation.new_scene("typing")
            for block_char in ["▉", "▓", "▒", "░"]:
                typing_scene.add_frame(block_char, 2, color=self.ciphertext_color)

            typing_scene.add_frame(random.choice(self.encrypted_symbols), 2, color=self.ciphertext_color)
            self.pending_chars.append(character)

    def prepare_data_for_decrypt_effect(self) -> None:
        """Prepares the data for the effect by building the animation for each character."""
        for character in self.terminal.characters:
            self.make_decrypting_animation_scenes(character)
            character.event_handler.register_event(
                EventHandler.Event.SCENE_COMPLETE,
                character.animation.scenes["fast_decrypt"],
                EventHandler.Action.ACTIVATE_SCENE,
                character.animation.scenes["slow_decrypt"],
            )
            character.event_handler.register_event(
                EventHandler.Event.SCENE_COMPLETE,
                character.animation.scenes["slow_decrypt"],
                EventHandler.Action.ACTIVATE_SCENE,
                character.animation.scenes["discovered"],
            )
            character.animation.activate_scene(character.animation.scenes["fast_decrypt"])
            self.animating_chars.append(character)

    def run(self) -> None:
        """Runs the effect."""
        self.prepare_data_for_type_effect()
        self.run_type_effect()
        self.prepare_data_for_decrypt_effect()
        self.run_decryption_effect()

    def run_type_effect(self) -> None:
        """Runs the typing out the characters effect."""
        self.terminal.print()
        while self.pending_chars or self.animating_chars:
            if self.pending_chars:
                if random.randint(0, 100) <= 75:
                    next_character = self.pending_chars.pop(0)
                    next_character.is_active = True
                    next_character.animation.activate_scene(next_character.animation.scenes["typing"])
                    self.animating_chars.append(next_character)
            self.animate_chars()
            # remove completed chars from animating chars
            self.animating_chars = [
                animating_char
                for animating_char in self.animating_chars
                if not animating_char.animation.active_scene_is_complete()
            ]
            self.terminal.print()

    def run_decryption_effect(self) -> None:
        while self.animating_chars:
            self.animate_chars()
            self.animating_chars = [
                animating_char
                for animating_char in self.animating_chars
                if not animating_char.animation.active_scene_is_complete()
            ]
            self.terminal.print()

    def animate_chars(self) -> None:
        """Animates the characters by calling the tween method and printing the characters to the terminal."""
        for animating_char in self.animating_chars:
            animating_char.animation.step_animation()



import argparse
import random

import terminaltexteffects.utils.argtypes as argtypes
from terminaltexteffects.base_character import EffectCharacter, EventHandler
from terminaltexteffects.utils.terminal import Terminal
from terminaltexteffects.utils import graphics, motion, argtypes


def add_arguments(subparsers: argparse._SubParsersAction) -> None:
    """Adds arguments to the subparser.

    Args:
        subparser (argparse._SubParsersAction): subparser to add arguments to
    """
    effect_parser = subparsers.add_parser(
        "vhstape",
        formatter_class=argtypes.CustomFormatter,
        help="Lines of characters glitch left and right and lose detail like an old VHS tape.",
        description="Lines of characters glitch left and right and lose detail like an old VHS tape.",
        epilog=f"""
Example: terminaltexteffects vhstape -a 0.01 --glitch-line-colors 00ff00 ff0000 0000ff --base-color ffffff --final-color ffffff --glitch-line-chance 0.05 --noise-chance 0.004 --total-glitch-time 1000""",
    )
    effect_parser.set_defaults(effect_class=VHSTapeEffect)
    effect_parser.add_argument(
        "-a",
        "--animation-rate",
        type=argtypes.valid_animationrate,
        default=0.01,
        help="Minimum time, in seconds, between animation steps. This value does not normally need to be modified. Use this to increase the playback speed of all aspects of the effect. This will have no impact beyond a certain lower threshold due to the processing speed of your device.",
    )
    effect_parser.add_argument(
        "--base-color",
        type=argtypes.valid_color,
        default="ffffff",
        metavar="(XTerm [0-255] OR RGB Hex [000000-ffffff])",
        help="Color for the characters when they are not glitching.",
    )
    effect_parser.add_argument(
        "--glitch-line-colors",
        type=argtypes.valid_color,
        nargs="*",
        default=["ffffff", "ff0000", "00ff00", "0000ff", "ffffff"],
        metavar="(XTerm [0-255] OR RGB Hex [000000-ffffff])",
        help="Space separated, unquoted, list of colors for the characters when a single line is glitching. Colors are applied in order as an animation.",
    )
    effect_parser.add_argument(
        "--glitch-wave-colors",
        type=argtypes.valid_color,
        nargs="*",
        default=["ffffff", "ff0000", "00ff00", "0000ff", "ffffff"],
        metavar="(XTerm [0-255] OR RGB Hex [000000-ffffff])",
        help="Space separated, unquoted, list of colors for the characters in lines that are part of the glitch wave. Colors are applied in order as an animation.",
    )
    effect_parser.add_argument(
        "--noise-colors",
        type=argtypes.valid_color,
        nargs="*",
        default=["1e1e1f", "3c3b3d", "6d6c70", "a2a1a6", "cbc9cf", "ffffff"],
        metavar="(XTerm [0-255] OR RGB Hex [000000-ffffff])",
        help="Space separated, unquoted, list of colors for the characters during the noise phase.",
    )
    effect_parser.add_argument(
        "--final-color",
        type=argtypes.valid_color,
        default="ffffff",
        metavar="(XTerm [0-255] OR RGB Hex [000000-ffffff])",
        help="Color for the final characters after the glitching has resolved.",
    )
    effect_parser.add_argument(
        "--glitch-line-chance",
        type=argtypes.valid_float_between_zero_and_one,
        default=0.05,
        metavar="(0.0-1.0)",
        help="Chance that a line will glitch on any given frame.",
    )
    effect_parser.add_argument(
        "--noise-chance",
        type=argtypes.valid_float_between_zero_and_one,
        default=0.004,
        metavar="(0.0-1.0)",
        help="Chance that all characters will experience noise on any given frame.",
    )
    effect_parser.add_argument(
        "--total-glitch-time",
        type=argtypes.valid_duration,
        default=1000,
        metavar="(int >= 1)",
        help="Total time, animation steps, that the glitching phase will last.",
    )


class Line:
    """
    Represents a line of characters with various effects.

    Args:
        characters (list[EffectCharacter]): List of EffectCharacter objects representing the characters in the line.
        args (argparse.Namespace): Namespace object containing command-line arguments.

    Attributes:
        characters (list[EffectCharacter]): List of EffectCharacter objects representing the characters in the line.
        args (argparse.Namespace): Namespace object containing command-line arguments.

    Methods:
        build_line_effects(): Builds the line effects for each character.
        snow(): Activates the snow animation for each character.
        set_hold_time(hold_time: int): Sets the hold time for glitch waypoints for each character.
        glitch(final=False): Activates the glitch animation for each character.
        restore(): Activates the restore animation for each character.
        activate_waypoint(waypoint_id: str): Activates a specific waypoint for each character.
        line_movement_complete(): Checks if the movement of all characters in the line is complete.

    """

    def __init__(self, characters: list[EffectCharacter], args: argparse.Namespace) -> None:
        self.characters = characters
        self.args = args
        self.build_line_effects()

    def build_line_effects(self) -> None:
        glitch_line_colors = self.args.glitch_line_colors
        snow_chars = ["#", "*", ".", ":"]
        noise_colors = self.args.noise_colors
        offset = random.randint(4, 25)
        direction = random.choice((-1, 1))
        hold_time = random.randint(1, 50)
        for character in self.characters:
            # make glitch and restore waypoints
            glitch_path = character.motion.new_path("glitch", speed=2, hold_time=hold_time)
            glitch_wpt = glitch_path.new_waypoint(
                "glitch", motion.Coord(character.input_coord.column + (offset * direction), character.input_coord.row)
            )
            restore_path = character.motion.new_path("restore", speed=2)
            restore_wpt = restore_path.new_waypoint("restore", character.input_coord)
            # make glitch wave waypoints
            glitch_wave_mid_path = character.motion.new_path("glitch_wave_mid", speed=2)
            glitch_wave_mid_wpt = glitch_wave_mid_path.new_waypoint(
                "glitch_wave_mid", motion.Coord(character.input_coord.column + 8, character.input_coord.row)
            )
            glitch_wave_end_path = character.motion.new_path("glitch_wave_end", speed=2)
            glitch_wave_end_wpt = glitch_wave_end_path.new_waypoint(
                "glitch_wave_end", motion.Coord(character.input_coord.column + 14, character.input_coord.row)
            )

            # make glitch scenes
            base_scn = character.animation.new_scene("base")
            base_scn.add_frame(character.input_symbol, duration=1, color=self.args.base_color)
            glitch_scn_forward = character.animation.new_scene("rgb_glitch_fwd", sync=graphics.SyncMetric.STEP)
            for color in glitch_line_colors:
                glitch_scn_forward.add_frame(character.input_symbol, duration=1, color=color)
            glitch_scn_backward = character.animation.new_scene("rgb_glitch_bwd", sync=graphics.SyncMetric.STEP)
            for color in glitch_line_colors[::-1]:
                glitch_scn_backward.add_frame(character.input_symbol, duration=1, color=color)
            snow_scn = character.animation.new_scene("snow")
            for _ in range(25):
                snow_scn.add_frame(random.choice(snow_chars), duration=2, color=random.choice(noise_colors))
            snow_scn.add_frame(character.input_symbol, duration=1, color=self.args.base_color)
            final_snow_scn = character.animation.new_scene("final_snow")
            final_redraw_scn = character.animation.new_scene("final_redraw")
            final_redraw_scn.add_frame("█", duration=10, color="ffffff")
            final_redraw_scn.add_frame(character.input_symbol, duration=1, color=self.args.final_color)

            for _ in range(50):
                final_snow_scn.add_frame(random.choice(snow_chars), duration=2, color=random.choice(noise_colors))
            # register events
            character.event_handler.register_event(
                EventHandler.Event.PATH_COMPLETE, glitch_path, EventHandler.Action.ACTIVATE_PATH, restore_path
            )
            character.event_handler.register_event(
                EventHandler.Event.PATH_ACTIVATED,
                glitch_path,
                EventHandler.Action.ACTIVATE_SCENE,
                glitch_scn_forward,
            )
            character.event_handler.register_event(
                EventHandler.Event.PATH_ACTIVATED,
                restore_path,
                EventHandler.Action.ACTIVATE_SCENE,
                glitch_scn_backward,
            )
            character.event_handler.register_event(
                EventHandler.Event.PATH_ACTIVATED,
                glitch_wave_mid_path,
                EventHandler.Action.ACTIVATE_SCENE,
                glitch_scn_forward,
            )
            character.event_handler.register_event(
                EventHandler.Event.PATH_ACTIVATED,
                glitch_wave_end_path,
                EventHandler.Action.ACTIVATE_SCENE,
                glitch_scn_forward,
            )
            character.event_handler.register_event(
                EventHandler.Event.SCENE_COMPLETE, glitch_scn_backward, EventHandler.Action.ACTIVATE_SCENE, base_scn
            )

    def snow(self) -> None:
        """
        Activates the snow animation for each character in the effect.
        """
        for character in self.characters:
            character.animation.activate_scene(character.animation.scenes["snow"])

    def set_hold_time(self, hold_time: int) -> None:
        """
        Set the hold time for each character's glitch motion.

        Args:
            hold_time (int): The hold time in animation steps.

        Returns:
            None
        """
        for character in self.characters:
            character.motion.paths["glitch"].hold_time = hold_time

    def glitch(self, final=False) -> None:
        """
        Apply glitch effect to the characters.

        Args:
            final (bool): If True, set hold_time to 0 for glitch and restore waypoints.

        Returns:
            None
        """
        for character in self.characters:
            glitch_path = character.motion.paths["glitch"]
            restore_path = character.motion.paths["restore"]
            if final:
                glitch_path.hold_time = 0
                restore_path.hold_time = 0
            glitch_path.speed = 40 / random.randint(20, 40)
            restore_path.speed = 40 / random.randint(20, 40)
            character.motion.activate_path(character.motion.paths["glitch"])

    def restore(self) -> None:
        for character in self.characters:
            restore_path = character.motion.paths["restore"]
            restore_path.speed = 40 / random.randint(20, 40)
            character.motion.activate_path(restore_path)

    def activate_path(self, path_id: str) -> None:
        for character in self.characters:
            character.motion.activate_path(character.motion.paths[path_id])

    def line_movement_complete(self):
        return all(character.motion.movement_is_complete() for character in self.characters)


class VHSTapeEffect:
    """
    Represents a VHS tape effect for terminal text.

    Attributes:
        terminal (Terminal): The terminal object.
        args (argparse.Namespace): The command-line arguments.
        pending_chars (list[EffectCharacter]): The list of pending effect characters.
        animating_chars (list[EffectCharacter]): The list of animating effect characters.
        lines (dict[int, Line]): The dictionary of lines indexed by row index.
        active_glitch_wave_top (int | None): The top row index of the active glitch wave, or None if no active wave.
        active_glitch_wave_lines (list[Line]): The list of lines in the active glitch wave.
        active_glitch_lines (list[Line]): The list of active glitch lines.
    """

    def __init__(self, terminal: Terminal, args: argparse.Namespace):
        self.terminal = terminal
        self.args = args
        self.pending_chars: list[EffectCharacter] = []
        self.animating_chars: list[EffectCharacter] = []
        self.lines: dict[int, Line] = {}
        self.active_glitch_wave_top: int | None = None
        self.active_glitch_wave_lines: list[Line] = []
        self.active_glitch_lines: list[Line] = []

    def prepare_data(self) -> None:
        for row_index, characters in self.terminal.input_by_row().items():
            self.lines[row_index] = Line(characters, self.args)
        for character in self.terminal.characters:
            character.is_active = True
            character.animation.activate_scene(character.animation.scenes["base"])

    def glitch_wave(self) -> None:
        if not self.active_glitch_wave_top:
            if self.terminal.output_area.top >= 3:
                # choose a wave top index in the top half of the output area or at least 3 rows up
                self.active_glitch_wave_top = random.randint(
                    max((3, round(self.terminal.output_area.top * 0.5))), self.terminal.output_area.top
                )
            else:
                # not enough room for a wave
                return

        # if all lines have completed movement, proceed to move/restore wave
        if all(line.line_movement_complete() for line in self.active_glitch_wave_lines):
            if self.active_glitch_wave_lines:
                # only move 30% of the time
                if random.random() < 0.3:
                    # if moving, only move up 10% of the time
                    if random.random() < 0.3:
                        wave_top_delta = 1
                    else:
                        wave_top_delta = -1
                else:
                    wave_top_delta = 0
                self.active_glitch_wave_top += wave_top_delta
                # clamp wave top to output area
                self.active_glitch_wave_top = max(2, min(self.active_glitch_wave_top, self.terminal.output_area.top))
            # get the lines for the wave
            new_wave_lines: list[Line] = []
            for line_index in range(self.active_glitch_wave_top - 2, self.active_glitch_wave_top + 1):
                if line_index in self.lines:
                    new_wave_lines.append(self.lines[line_index])

            # restore any lines that are no longer part of the wave
            for line in self.active_glitch_wave_lines:
                if line not in new_wave_lines:
                    line.restore()
                    self.animating_chars.extend(line.characters)
            self.active_glitch_wave_lines = new_wave_lines

            if self.active_glitch_wave_top < 3:
                # wave at bottom, restore lines
                for line in self.active_glitch_wave_lines:
                    line.restore()
                    self.animating_chars.extend(line.characters)
                self.active_glitch_wave_top = None
                self.active_glitch_wave_lines = []

            else:
                for line, path_id in zip(
                    self.active_glitch_wave_lines, ("glitch_wave_mid", "glitch_wave_end", "glitch_wave_mid")
                ):
                    line.activate_path(path_id)
                    self.animating_chars.extend(line.characters)

    def run(self) -> None:
        """Runs the effect."""
        self.prepare_data()

        self.terminal.print()
        glitching_steps_elapsed = 0
        phase = "glitching"
        to_redraw = list(self.lines.values())
        redrawing = False
        while phase != "complete" or self.animating_chars:
            if phase == "glitching":
                # Check if all active glitch wave lines have completed their movement, if so move the wave
                if not self.active_glitch_wave_lines or all(
                    line.line_movement_complete() for line in self.active_glitch_wave_lines
                ):
                    self.glitch_wave()
                # Remove completed glitch lines from active glitch lines
                self.active_glitch_lines = [
                    line for line in self.active_glitch_lines if not line.line_movement_complete()
                ]
                # Randomly add new glitch lines
                if random.random() < self.args.glitch_line_chance and len(self.active_glitch_lines) < 3:
                    glitch_line: Line = random.choice(list(self.lines.values()))
                    if glitch_line not in self.active_glitch_wave_lines and glitch_line not in self.active_glitch_lines:
                        glitch_line.set_hold_time(random.randint(30, 120))
                        self.active_glitch_lines.append(glitch_line)
                        glitch_line.glitch()
                        self.animating_chars.extend(glitch_line.characters)
                # Randomly add noise to all lines
                if random.random() < self.args.noise_chance:
                    for line in self.lines.values():
                        line.snow()
                        if line not in self.active_glitch_wave_lines and line not in self.active_glitch_lines:
                            self.animating_chars.extend(line.characters)
                glitching_steps_elapsed += 1
                # Check if glitching time has reached the total glitch time
                if glitching_steps_elapsed >= self.args.total_glitch_time:
                    # Restore glitch wave lines
                    for line in self.active_glitch_wave_lines:
                        line.restore()
                    # Restore glitch lines
                    for line in self.active_glitch_lines:
                        line.restore()
                    phase = "noise"

            elif phase == "noise":
                # Activate final snow animation for all characters
                if not self.animating_chars:
                    for character in self.terminal.characters:
                        character.animation.activate_scene(character.animation.scenes["final_snow"])
                        self.animating_chars.append(character)
                    phase = "redraw"

            elif phase == "redraw":
                # Redraw lines one by one
                if redrawing or not self.animating_chars:
                    redrawing = True
                    if to_redraw:
                        next_line = to_redraw.pop()
                        for character in next_line.characters:
                            character.animation.activate_scene(character.animation.scenes["final_redraw"])
                            self.animating_chars.append(character)
                    else:
                        phase = "complete"
            self.animate_chars()
            self.terminal.print()

            # Remove completed characters from animating_chars list
            self.animating_chars = [
                animating_char
                for animating_char in self.animating_chars
                if not animating_char.animation.active_scene_is_complete()
                or not animating_char.motion.movement_is_complete()
            ]

    def animate_chars(self) -> None:
        """Animates the characters by calling the move method and step animation. Move characters prior to stepping animation
        to ensure waypoint synced animations have the latest waypoint progress information."""
        for animating_char in self.animating_chars:
            animating_char.motion.move()
            animating_char.animation.step_animation()
